"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const path = require("path");
const url = require("url");
function buildLD(schemabase) {
    const _label = (jsd) => path.parse(new url.URL(jsd.title).pathname).name;
    const datatypeLD = (jsd) => {
        const label = _label(jsd);
        return {
            '@type': 'rdfs:Datatype',
            '@id': `sdo:${label}`,
            'rdfs:label': label,
            'rdfs:comment': jsd.description,
        };
    };
    const classLD = (jsd, propertybase) => {
        const label = _label(jsd);
        return {
            '@type': 'rdfs:Class',
            '@id': `sdo:${label}`,
            'rdfs:label': label,
            'rdfs:comment': jsd.description,
            'rdfs:subClassOf': (label !== 'Thing') ? { '@id': `sdo:${path.parse(jsd.allOf[0].$ref).name}` } : null,
            'rdfs:member': Object.entries(jsd.allOf[1].properties).map((entry) => {
                const prop_name = entry[0];
                if (propertybase.find((sch) => sch.title === `http://schema.org/${prop_name}`))
                    return { '@id': `sdo:${prop_name}` };
                else
                    throw new ReferenceError(`No corresponding jsd file was found for member subschema \`${label}#${prop_name}\`.`);
            }),
            superClassOf: [],
            valueOf: [],
        };
    };
    const propertyLD = (jsd) => {
        const label = _label(jsd);
        return {
            '@type': 'rdf:Property',
            '@id': `sdo:${label}`,
            'rdfs:label': label,
            'rdfs:comment': jsd.description,
            'rdfs:subPropertyOf': (jsd.allOf[0] !== true) ? { '@id': `sdo:${path.parse(jsd.allOf[0].$ref).name.split('.')[0]}` } : null,
            'rdfs:domain': [],
            'rdfs:range': jsd.definitions.ExpectedType.anyOf.map((schema) => ({ '@id': `sdo:${path.parse(schema.$ref).name}` })),
            superPropertyOf: [],
            $rangeArray: jsd.allOf[1].anyOf.length === 2,
        };
    };
    const processSubclasses = (jsonld, classbase) => {
        const superclass = jsonld['rdfs:subClassOf'];
        const referenced = superclass && classbase.find((c) => c['@id'] === superclass['@id']) || null;
        if (referenced) {
            referenced.superClassOf.push({ '@id': jsonld['@id'] });
        }
    };
    const processSubproperties = (jsonld, propertybase) => {
        const superproperty = jsonld['rdfs:subPropertyOf'];
        const referenced = superproperty && propertybase.find((p) => p['@id'] === superproperty['@id']) || null;
        if (referenced) {
            referenced.superPropertyOf.push({ '@id': jsonld['@id'] });
        }
    };
    const processDomains = (jsonld, propertybase) => {
        jsonld['rdfs:member'].forEach((property_ref) => {
            const referenced = propertybase.find((p) => p['@id'] === property_ref['@id']) || null;
            if (referenced) {
                referenced['rdfs:domain'].push({ '@id': jsonld['@id'] });
            }
        });
    };
    const processValueOf = (jsonld, classbase) => {
        jsonld['rdfs:range'].forEach((class_ref) => {
            const referenced = classbase.find((c) => c['@id'] === class_ref['@id']) || null;
            if (referenced) {
                referenced.valueOf.push({ '@id': jsonld['@id'] });
            }
        });
    };
    const datatypeJSDs = schemabase.filter((jsd) => jsd.$schema === 'http://json-schema.org/draft-07/schema#');
    const classJSDs = schemabase.filter((jsd) => jsd.$schema === 'https://chharvey.github.io/schemaorg-jsd/meta/type.jsd#');
    const propertyJSDs = schemabase.filter((jsd) => jsd.$schema === 'https://chharvey.github.io/schemaorg-jsd/meta/member.jsd#');
    const datatypeLDs = datatypeJSDs.map((jsd) => datatypeLD(jsd));
    const classLDs = classJSDs.map((jsd) => classLD(jsd, propertyJSDs));
    const propertyLDs = propertyJSDs.map((jsd) => propertyLD(jsd));
    classLDs.forEach((jsonld) => {
        processSubclasses(jsonld, classLDs);
        processDomains(jsonld, propertyLDs);
    });
    propertyLDs.forEach((jsonld) => {
        processSubproperties(jsonld, propertyLDs);
        processValueOf(jsonld, classLDs);
    });
    return {
        '@context': {
            sdo: 'http://schema.org/',
            rdf: 'https://www.w3.org/1999/02/22-rdf-syntax-ns#',
            rdfs: 'http://www.w3.org/2000/01/rdf-schema#',
            superClassOf: { '@reverse': 'rdfs:subClassOf' },
            superPropertyOf: { '@reverse': 'rdfs:subPropertyOf' },
            valueOf: { '@reverse': 'rdfs:range' },
        },
        '@graph': [
            ...datatypeLDs,
            ...classLDs,
            ...propertyLDs,
        ],
    };
}
exports.buildLD = buildLD;
exports.buildTS = (jsonlddocument) => {
    const _linklist = (lds) => lds.map((obj) => ` * - {@link ${obj['@id'].split(':')[1]}}`).join('\n');
    const datatypeTS = (ld) => {
        let SDODatatypeNames;
        (function (SDODatatypeNames) {
            SDODatatypeNames[SDODatatypeNames["Boolean"] = 0] = "Boolean";
            SDODatatypeNames[SDODatatypeNames["Date"] = 1] = "Date";
            SDODatatypeNames[SDODatatypeNames["DateTime"] = 2] = "DateTime";
            SDODatatypeNames[SDODatatypeNames["Integer"] = 3] = "Integer";
            SDODatatypeNames[SDODatatypeNames["Number"] = 4] = "Number";
            SDODatatypeNames[SDODatatypeNames["Text"] = 5] = "Text";
            SDODatatypeNames[SDODatatypeNames["Time"] = 6] = "Time";
            SDODatatypeNames[SDODatatypeNames["URL"] = 7] = "URL";
        })(SDODatatypeNames || (SDODatatypeNames = {}));
        const alias = ld['rdfs:label'];
        const type = ({
            Boolean: 'boolean',
            Date: 'string',
            DateTime: 'string',
            Integer: 'number',
            Number: 'number',
            Text: 'string',
            Time: 'string',
            URL: 'string',
        })[alias];
        return `
			/**
			 * ${ld['rdfs:comment']}
			 *
			 * @see http://schema.org/${ld['rdfs:label']}
			 */
			export type ${alias} = ${type}
		`.replace(/\n\t\t\t/g, '\n');
    };
    const classTS = (ld) => `
		/**
		 * ${ld['rdfs:comment']}
		 *
		 * ${(ld['superClassOf'].length) ? `*(Non-Normative):* Known subclasses:\n${_linklist(ld['superClassOf'])}\n` : ''}
		 * ${(ld['valueOf'].length) ? `*(Non-Normative):* May appear as values of:\n${_linklist(ld['valueOf']).replace(/}/g, '_type}')}\n` : ''}
		 * @see http://schema.org/${ld['rdfs:label']}
		 */
		export interface ${ld['rdfs:label']} extends ${(ld['rdfs:subClassOf']) ? ld['rdfs:subClassOf']['@id'].split(':')[1] : 'JSONLDObject'} {
			${ld['rdfs:member'].map((member) => member['@id'].split(':')[1]).map((name) => `
				${name}?: ${name}_type
			`).join('')}
		}
	`.replace(/\n\t\t/g, '\n');
    const propertyTS = (ld) => {
        const rangeunion = `${ld['rdfs:range'].map((cls) => cls['@id'].split(':')[1]).join('|')}`;
        return `
			/**
			 * ${ld['rdfs:comment']}
			 *
			 * ${(ld['rdfs:subPropertyOf']) ? `Extends {@link ${ld['rdfs:subPropertyOf']['@id'].split(':')[1]}}` : ''}
			 * ${(ld['superPropertyOf'].length) ? `*(Non-Normative):* Known subproperties:\n${_linklist(ld['superPropertyOf'])}\n` : ''}
			 * ${(ld['rdfs:domain'].length) ? `*(Non-Normative):* Property of:\n${_linklist(ld['rdfs:domain'])}\n` : ''}
			 * @see http://schema.org/${ld['rdfs:label']}
			 */
			type ${ld['rdfs:label']}_type = ${rangeunion}${(ld['$rangeArray']) ? `|(${rangeunion})[]` : ''}
		`.replace(/\n\t\t\t/g, '\n');
    };
    const JSONLD = jsonlddocument['@graph'];
    return [
        `import { JSONLDObject } from '@chharvey/requirejson'`,
        ...JSONLD.filter((jsonld) => jsonld['@type'] === 'rdfs:Datatype').map((ld) => datatypeTS(ld)),
        ...JSONLD.filter((jsonld) => jsonld['@type'] === 'rdfs:Class').map((ld) => classTS(ld)),
        ...JSONLD.filter((jsonld) => jsonld['@type'] === 'rdf:Property').map((ld) => propertyTS(ld)),
    ].join('');
};
